/*
 * OCSP responder - Massimiliano Pala (madwolf@openca.org)
 * OpenCA Licensed Software
 *
 * Copyright (c) 2001-2006 The OpenCA Project.  All rights reserved.
 *
 * ====================================================================
 *
 * This product includes cryptographic software written by Eric Young
 * (eay@cryptsoft.com).  This product includes software written by Tim
 * Hudson (tjh@cryptsoft.com).
 *
 */

#include "general.h"

#include <openssl/asn1.h>
#include <openssl/txt_db.h>
#include <openssl/conf.h>
#include <openssl/buffer.h>
#include <openssl/x509.h>

#include "crypto.h"

extern OCSPD_CONFIG *ocspd_conf;

char * ocspd_parse_cdp_ext ( STACK_OF(X509) *sk_cert ) {

	STACK_OF(DIST_POINT) *sk_cdp = NULL;
	DIST_POINT *cdp = NULL;

	X509 *cert = NULL;

	STACK_OF(CONF_VALUE) *sk_val = NULL;
	CONF_VALUE *v = NULL;

	char *ret = NULL;
	int k = -1;

	if( (!sk_cert) || (sk_X509_num(sk_cert) < 1) ) {
		return NULL;
	};

	if( (cert = sk_X509_value(sk_cert, 0)) == NULL ) {
		return NULL;
	}

	if(( sk_cdp=X509_get_ext_d2i(cert, NID_crl_distribution_points, 
						NULL, NULL)) == NULL ) {
		return NULL;
	}

	/* Should we go through the whole stack ? Maybe, now we just
	   take the first value... */
	if(( cdp = sk_DIST_POINT_pop( sk_cdp )) == NULL ) {
		sk_DIST_POINT_free( sk_cdp );
		return NULL;
	}

	if( cdp->distpoint ) {
		if(cdp->distpoint->type == 0) {
			if( cdp->distpoint->name.fullname ) {
				sk_val = i2v_GENERAL_NAMES(NULL,
                                         cdp->distpoint->name.fullname, 
						sk_val);
				k=0;
				for( ;; ) {
					v = sk_CONF_VALUE_value( sk_val, k++ );
					if( v == NULL ) break;
					
					if( strncmp("URI", v->name, 3) == 0 ) {
						if( ocspd_conf->verbose )
							syslog(LOG_INFO,
								"INFO::Found CDP in cert %s:%s", v->name, v->value );
						ret = strdup( v->value );
						break;
					}
				}
				sk_CONF_VALUE_free(sk_val);
			}
		} else {
			DIST_POINT_free( cdp );
			sk_DIST_POINT_free( sk_cdp );
		}
	}

	return ret;

}

